# -*- coding: utf-8 -*-
# Copyright (C) 2016-2017 Cordell Bloor
# Published under the MIT License

"""Shared utilities between checkguard, guard2once and once2guard."""

from __future__ import print_function
import sys
import os
import re
from fnmatch import fnmatch
from functools import partial
from itertools import islice

def next_nonempty_line(contents, start):
    """
    Returns the index of the first character of the next line containing
    non-whitespace characters, or returns len(contents) if there is no such
    line.
    """
    line_end = len(contents) - 1
    for i, c in enumerate(islice(contents, start, None), start):
        if c == '\n':
           line_end = i
        elif c not in ' \t':
           break
    return line_end + 1

def next_line(contents, start):
    """
    Returns the index of the first character of the next line, or returns
    len(contents) if there is no next line.
    """
    for i, c in enumerate(islice(contents, start, None), start):
        if c == '\n':
           return i + 1
    return len(contents)

def guess_guard(contents):
    """
    Returns the guard, as well as the start and end indexes of the include
    guard from the start of the file, or throws ValueError if not found.
    Comments are not supported.
    """
    regex = re.compile(r"^[ \t]*\#[ \t]*ifndef[ \t]+([\w]+)"
        + r"[ \t]*\n[ \t]*\#[ \t]*define[ \t]+\1([ \t]+1)?[ \t]*$",
        re.MULTILINE)
    guard_found = False
    for match in regex.finditer(contents):
        # if it matches our regex, it's probably a guard, but check
        # that there's some content within to be a little more sure
        nextline_start = next_nonempty_line(contents, match.end())
        nextline_end = next_line(contents, nextline_start)
        try:
            index_guard_end(contents, nextline_start, nextline_end)
        except ValueError as e:
            guard_found = True
            break
    if not guard_found:
        raise ValueError('guard start not found')
    return (match.group(1),) + match.span()

def index_pragma_once(contents):
    """
    Returns the start and end indexes of the pragma once directive from
    the start of the file, or throws ValueError if not found. Comments
    are not supported.
    """
    regex = re.compile(r"^[ \t]*\#[ \t]*pragma[ \t]+once[ \t]*$",
        re.MULTILINE)
    match = regex.search(contents)
    if not match:
        raise ValueError('pragma once not found')
    return match.span()

def index_guard_start(contents, guard_symbol):
    """
    Returns the start and end indexes of the include guard from the start of the
    file, or throws ValueError if not found. Comments are not supported.
    """
    regex = re.compile(r"^[ \t]*\#[ \t]*ifndef[ \t]+" + guard_symbol
        + r"[ \t]*\n[ \t]*\#[ \t]*define[ \t]+" + guard_symbol
        + r"([ \t]+1)?[ \t]*$",
        re.MULTILINE)
    match = regex.search(contents)
    if not match:
        raise ValueError('guard start not found')
    return match.span()

def index_guard_end(contents, pos=0, endpos=None):
    """
    Returns the start and end indexes of the last endif line from the
    file, or throws ValueError if not found. Comments are not supported.
    """
    if endpos is None:
        endpos = len(contents)
    regex = re.compile(r"^[ \t]*\#[ \t]*endif([/ \t]+.*|[ \t]*)$",
        re.MULTILINE)
    match = None
    for match in regex.finditer(contents, pos, endpos):
        pass
    if not match:
        raise ValueError('guard end not found')
    return match.span()

def get_file_contents(filename):
    with open(filename, 'r') as f:
        return f.read()

def write_file_contents(filename, contents):
    with open(filename, 'w') as f:
        f.write(contents)

def print_error(error):
    print(error, file=sys.stderr)

def is_header_file(filename):
    '''Returns true if the given file is identified as a C/C++ header file.'''
    return filename.endswith(('.h', '.hpp', '.H', '.hh'))

def is_excluded(filepath, exclusions):
    '''Returns true if the given file matches any of the exclusion patterns.'''
    return any(map(partial(fnmatch, filepath), exclusions))

def apply_to_headers(func, directory, exclusions):
    for root, dirs, files in os.walk(directory, onerror=print_error):
        for filename in files:
            filepath = os.path.join(root, filename)
            if is_header_file(filename) and not is_excluded(filepath, exclusions):
                func(filepath, filename)



# -*- coding: utf-8 -*-
# Copyright (C) 2016-2017 Cordell Bloor
# Published under the MIT License

"""Find C or C++ header files with incorrect or missing include guards."""

from __future__ import print_function
import argparse
import sys
import os
from functools import partial
from .pattern_compiler import compile_pattern, ParserError
from .util import (guess_guard, index_guard_start, index_guard_end,
    index_pragma_once, get_file_contents, apply_to_headers)

__version__ = "2.1.0"

def is_protected_by_guard(contents, guard_symbol):
    try:
        if guard_symbol:
            index_guard_start(contents, guard_symbol)
        else:
            guess_guard(contents)

        index_guard_end(contents)
        return True
    except ValueError:
        return False

def is_protected_by_once(contents):
    try:
        index_pragma_once(contents)
        return True
    except ValueError:
        return False

def is_protected(contents, options):
    return (options.accept_guard and is_protected_by_guard(contents, options.guard)
        or options.accept_once and is_protected_by_once(contents))

def is_file_protected(filename, options):
    contents = get_file_contents(filename)
    return is_protected(contents, options)

def process_file(filepath, filename, options):
    class Context:
        pass
    ctx = Context()
    ctx.filepath = filepath
    ctx.filename = filename

    options.guard = options.create_guard(ctx)

    try:
        if not is_file_protected(filepath, options):
            print(filepath)
    except Exception as e:
        print(e, file=sys.stderr)

def process_pattern(guard_pattern):
    create_guard = lambda ctx: None
    if guard_pattern is not None:
        try:
            create_guard = compile_pattern(guard_pattern)
        except ParserError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    return create_guard

def main():
    parser = argparse.ArgumentParser(
            description='Find C or C++ header files with incorrect or missing '
            'include guards.')
    parser.add_argument('files',
            metavar='file',
            nargs='+',
            help='the file(s) to check; directories require the recursive '
            'option')
    parser.add_argument('-V','--version', action='version',
            version='%(prog)s ' + __version__)
    parser.add_argument('-r','--recursive',
            action='store_true',
            dest='recursive',
            help='recursively search directories for headers')
    parser.add_argument('-p','--pattern',
            default=None,
            metavar='pattern',
            help='check that include guards match the specified pattern. For '
            "example, -p 'name | upper' would create an expectation that "
            'Match.h has the include guard MATCH_H. See the docs on GitHub '
            'for a full description of the guard pattern language.')
    parser.add_argument('-e','--exclude',
            action='append',
            dest='exclusions',
            metavar='pattern',
            default=[],
            help='exclude files that match the given fnmatch pattern. '
            'Any * is a wildcard matching everything; '
            'a ? matches any single character; '
            '[_] matches any characters within the brackets; '
            'and [!_] matches any characters not within the brackets.')
    parser.add_argument('-o','--only',
            dest='type',
            metavar='type',
            default='any',
            choices=['guard','once','g','o'],
            help='only accept the specified type of include protection. '
            "Use 'guard' or 'g' to only accept include guards, or "
            "use 'once' or 'o' to only accept #pragma once.")
    args = parser.parse_args()

    class Options:
        pass
    options = Options()
    options.accept_guard = args.type in ['g', 'guard', 'any']
    options.accept_once = args.type in ['o', 'once', 'any']
    options.create_guard = process_pattern(args.pattern)

    for f in args.files:
        if os.path.isdir(f):
            if args.recursive:
                process = partial(process_file, options=options)
                apply_to_headers(process, f, args.exclusions)
            else:
                print('"%s" is a directory' % f, file=sys.stderr)
                sys.exit(1)
        else:
            process_file(f, os.path.basename(f), options)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Cordell Bloor
# Published under the MIT License

"""Create functions that generate include guard tokens from patterns."""

import os
import re
from string import capwords

class ParseState:
    Normal, Token, SingleQuote, DoubleQuote, Complete = range(5)

class ParserError(Exception):
    pass

def next_token(pattern, start):
    """
    Given a pattern string and a start index, returns the end index of the next
    token, including whitespace on either side.
    """
    state = ParseState.Normal
    index = start
    for i in range(start, len(pattern)):
        if state == ParseState.SingleQuote:
            if pattern[i] == "'":
                state = ParseState.Complete
        elif state == ParseState.DoubleQuote:
            if pattern[i] == '"':
                state = ParseState.Complete
        else:
            if pattern[i].isspace():
                if state == ParseState.Token:
                    state = ParseState.Complete
            elif pattern[i] == '|':
                break # delimits tokens
            elif pattern[i] == '"':
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.DoubleQuote
            elif pattern[i] == "'":
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.SingleQuote
            else:
                if state == ParseState.Complete:
                    break # start of new token
                state = ParseState.Token
        index = i

    if state in [ParseState.SingleQuote, ParseState.DoubleQuote]:
        raise ParserError('Pattern contains unclosed quote')

    return index + 1

def unquote(s):
    """
    Strips matching single and double quotes from the start and end of the
    given string.
    """
    if len(s) > 1 and ((s[0] == '"' and s[-1] == '"') or
                       (s[0] == "'" and s[-1] == "'")):
        return s[1:-1]
    else:
        return s

def tokenize(pattern):
    tokens = []
    start = 0
    end = 0
    while end != len(pattern):
        end = next_token(pattern, start)
        tokens.append(pattern[start:end].strip())
        start = end
    return tokens

def splitpath(filepath, crumbs):
    """
    Splits a file path on the directory seperator
    """
    idx = len(filepath)
    for i in range(crumbs + 1):
        if idx >= 0:
            idx = filepath.rfind(os.sep, 0, idx)
    if idx <= 0:
        return filepath
    else:
        return filepath[idx+1:]

def snake(s):
    """
    Converts an input string in PascalCase to snake_case.
    """
    snek = []
    prev_up = False
    prev_alnum = False
    for idx, c in enumerate(s):
        alnum = c.isalnum()
        up = c.isupper()
        next_up = s[idx+1].isupper() if idx+1 < len(s) else False
        if (up and not prev_up and prev_alnum and idx != 0 or
            up and prev_up and not next_up and idx != len(s)):
            snek.append('_')
        snek.append(c.lower())
        prev_alnum = alnum
        prev_up = up
    return ''.join(snek)

def pascal(s):
    """
    Converts an input string in snake_case to PascalCase. Also handles
    names containing spaces... god help you if you actually need that.
    """
    return capwords(s.replace('_',' ')).replace(' ','')

def sanitize(s):
    """
    Removes characters that are not allowed in macro names. Anything
    that's not alphanumeric is replaced with underscore.
    """
    return re.sub(r"\W", '_', s)

class Args:
    """
    An enum representing all arguments that can be passed to functions.
    """
    (Replace, ReplaceWith, AppendWith, PrependWith, SurroundWith,
        PathCrumbs, Remove) = range(1,8)

def compile_pattern(pattern):
    """
    Takes a pattern specification as a string, and returns a function that
    turns a file context into an include guard.
    """
    sources = ['name','path']

    chain = []
    function = None
    expected_arg = None
    optional_arg = False
    args = []
    raw = False

    tokens = tokenize(pattern)
    if not tokens:
        raise ParserError('Pattern is empty')
    elif tokens[0] not in sources:
        raise ParserError('First function in pattern must be a source')
    elif tokens[-1] == '|':
        raise ParserError('Pattern ends with pipe to nowhere')

    for token in tokens:
        if raw:
            raise ParserError('Raw function is a sink and must be last')

        if not expected_arg:
            if not function:
                function = token
            elif token == '|':
                function = None
            else:
                raise ParserError('Unexpected argument "%s" in pattern' % token)

            if token == 'name':
                chain.append(lambda ctx, s: ctx.filename)
            elif token == 'path':
                expected_arg = Args.PathCrumbs
                optional_arg = True
            elif token == 'upper':
                chain.append(lambda ctx, s: s.upper())
            elif token == 'lower':
                chain.append(lambda ctx, s: s.lower())
            elif token == 'snake':
                chain.append(lambda ctx, s: snake(s))
            elif token == 'pascal':
                chain.append(lambda ctx, s: pascal(s))
            elif token == 'remove':
                expected_arg = Args.Remove
            elif token == 'replace':
                expected_arg = Args.Replace
            elif token == 'append':
                expected_arg = Args.AppendWith
            elif token == 'prepend':
                expected_arg = Args.PrependWith
            elif token == 'surround':
                expected_arg = Args.SurroundWith
            elif token == 'raw':
                raw = True
            elif token != '|':
                raise ParserError('Unknown function "%s" in pattern' % token)
        elif token == '|':
            raise ParserError('Missing argument from "%s" in pattern' % function)
        elif expected_arg == Args.Remove:
            chain.append(lambda ctx, s, t=unquote(token): s.replace(t, ''))
            expected_arg = None
        elif expected_arg == Args.Replace:
            expected_arg = Args.ReplaceWith
            args.append(unquote(token))
        elif expected_arg == Args.ReplaceWith:
            chain.append(lambda ctx, s, f=args[0], t=unquote(token): s.replace(f, t))
            expected_arg = None
            args = []
        elif expected_arg == Args.AppendWith:
            chain.append(lambda ctx, s, suffix=unquote(token): s + suffix)
            expected_arg = None
        elif expected_arg == Args.PrependWith:
            chain.append(lambda ctx, s, prefix=unquote(token): prefix + s)
            expected_arg = None
        elif expected_arg == Args.SurroundWith:
            chain.append(lambda ctx, s, arg=unquote(token): arg + s + arg)
            expected_arg = None
        elif expected_arg == Args.PathCrumbs:
            try:
                crumbs = int(token)
            except ValueError:
                raise ParserError('Invalid argument "%s" given to path; must be an integer' % token)
            chain.append(lambda ctx, s, crumbs=crumbs: splitpath(ctx.filepath, crumbs))
            expected_arg = None
            optional_arg = False

    if optional_arg:
        if expected_arg == Args.PathCrumbs:
            chain.append(lambda ctx, s: ctx.filepath)
    elif expected_arg:
        raise ParserError('Missing argument from "%s" in pattern' % function)

    def process(ctx):
        s = ''
        for fn in chain:
            s = fn(ctx, s)
        return s if raw else sanitize(s)
    return process



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Cordell Bloor
# Published under the MIT License

"""Replace C and C++ include guards with #pragma once."""

from __future__ import print_function
import argparse
import sys
import os
from functools import partial
from .pattern_compiler import compile_pattern, ParserError
from .util import (guess_guard, index_guard_start, index_guard_end,
    get_file_contents, write_file_contents, apply_to_headers)

__version__ = "2.1.0"

def replace_guard(contents, guard, strip=False):
    try:
        if guard:
            start1, end1 = index_guard_start(contents, guard)
        else:
            guard, start1, end1 = guess_guard(contents)
        start2, end2 = index_guard_end(contents)
        if end2 < len(contents) and contents[end2] == '\n':
            end2 += 1

        replaced = (contents[:start1] + '#pragma once' +
                contents[end1:start2] + contents[end2:])

        if strip:
            return replaced.rstrip() + '\n'
        else:
            return replaced
    except ValueError:
        return None

def process_file(filepath, filename, options):
    class Context:
        pass
    ctx = Context()
    ctx.filepath = filepath
    ctx.filename = filename

    options.guard = options.create_guard(ctx)

    try:
        contents = get_file_contents(filepath)
        new_contents = replace_guard(contents, options.guard,
                strip=options.strip)
        if new_contents:
            write_file_contents(filepath, new_contents)
    except Exception as e:
        print(e, file=sys.stderr)

def process_guard_pattern(pattern):
    create_guard = lambda ctx: None
    if pattern is not None:
        try:
            create_guard = compile_pattern(pattern)
        except ParserError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    return create_guard

def main():
    parser = argparse.ArgumentParser(
            description='Replace C and C++ include guards with #pragma once.')
    parser.add_argument('files',
            metavar='file',
            nargs='+',
            help='the file(s) to check; directories require the recursive '
            'option')
    parser.add_argument('-V','--version', action='version',
            version='%(prog)s ' + __version__)
    parser.add_argument('-r','--recursive',
            action='store_true',
            dest='recursive',
            help='recursively search directories for headers')
    parser.add_argument('-p','--pattern',
            default=None,
            metavar='pattern',
            help='replace only include guards that match the specified '
            "pattern. For example, -p 'name | upper' would only change the "
            'guard in Match.h if it had the symbol MATCH_H. See the docs on '
            'GitHub for a full description of the guard pattern language.')
    parser.add_argument('-e','--exclude',
            action='append',
            dest='exclusions',
            metavar='pattern',
            default=[],
            help='exclude files that match the given fnmatch pattern. '
            'Any * is a wildcard matching everything; '
            'a ? matches any single character; '
            '[_] matches any characters within the brackets; '
            'and [!_] matches any characters not within the brackets.')
    parser.add_argument('-s','--strip',
            action='store_true',
            dest='strip',
            help='strip trailing whitespace after removing guard')
    args = parser.parse_args()

    class Options:
        pass
    options = Options()
    options.strip = args.strip
    options.create_guard = process_guard_pattern(args.pattern)

    for f in args.files:
        if os.path.isdir(f):
            if args.recursive:
                process = partial(process_file, options=options)
                apply_to_headers(process, f, args.exclusions)
            else:
                print('"%s" is a directory' % f, file=sys.stderr)
                sys.exit(1)
        else:
            process_file(f, os.path.basename(f), options)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name = 'guardonce',
    version = '2.1.0',
    description = 'Utilities for converting from C/C++ include guards to #pragma once and back again',
    license = 'MIT',
    url = 'https://github.com/cgmb/guardonce',
    author = 'Cordell Bloor',
    author_email = 'slavik81@gmail.com',
    packages = ['guardonce'],
    entry_points = {
        'console_scripts': [
            'checkguard = guardonce.checkguard:main',
            'guard2once = guardonce.guard2once:main',
            'once2guard = guardonce.once2guard:main',
        ]
    },
    extras_require = { 'test': ['nose'] },
    install_requires = [
    ],
    tests_require = ['guardonce[test]'],
    test_suite = "nose.collector",
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    long_description = """
Why Convert?
------------

Include guards suck. They're tiring to type and tedius to update. Worse,
the task of updating boilerplate leaves room for copy/paste errors, or
other mistakes. #pragma once is simpler and less error prone. That's why
you should convert to #pragma once.

Alas, though #pragma once is available on all the most commonly used
compilers, it's not available on *every* compiler. Perhaps one day
you'll add support for a platform with a barebones compiler with no
support for #pragma once and you'll have to convert back. That's ok.
It's easy!

What exactly is guardonce?
--------------------------

There are three main tools provided by guardonce:

1. ``checkguard`` helps find any broken include guards you may already
   have in your project. These should be addressed before converting.

2. ``guard2once`` converts files with include guards into files with
   #pragma once directives. This ensures your entire project is
   consistently using #pragma once.

3. ``once2guard`` converts files with #pragma once directives back into
   files with include guards. This ensures your entire project is
   consistently using include guards.

How to use:
-----------

First, check your project for broken headers. To recursively search your
project directories for the names of all files that lack proper include
guards, use the following command, substituting your project's directory
for the quoted string:

``checkguard -r "source_directory"``

By default, checkguard is very forgiving. It accepts either #pragma once
or anything that looks like an include guard. If you know that all your
guards should match some format, you can be more strict by using ``-p``
to specify `a pattern <docs/PatternLanguage.md>`__ to check against.

If certain files are not supposed to have include guards, feel free to
leave them be. Files without include guards are ignored by this next
step.

Now, all that remains is converting the headers to use #pragma once:

``guard2once -r "source_directory"``

You're done! Double check that the result matches your expectations and
start using #pragma once in your new code. Know that if you ever need to
switch back, it's as simple as:

``once2guard -r "source_directory"``
"""
)
